from tenji.request.request_base import RequestBase

class BuyItemRequest(RequestBase):
    def __init__(self, id: int, jan: int = None, users: bool = False) -> None:
        self.id = id
        self.jan = jan
        self.users = users

    def getPath(self) -> str:
        return f"{self.BASE_URL}item/{self.id}"

    def getMethod(self):
        return "POST"

    def getParams(self):
        p = {
            "commit": "loadWindow",
            "window": "buyItem",
            "soldBy": "users" if self.users else None,
            "jan": self.jan # used for user listings
        }
        return self.build_params(p)


from tenji.request.request_base import RequestBase


class ProfileRequest(RequestBase):
    def __init__(self, username: str) -> None:
        self.username = username

    def getPath(self) -> str:
        return f"{self.BASE_URL}profile/{self.username}"

    def getMethod(self):
        return "GET"



from bs4 import PageElement
from tenji.model.shop.shop import Shop
from tenji.parser.parser_base import ParserBase


class ShopParser(ParserBase):
    def get_shop_field_element(self, label: str) -> PageElement():
        label_node = self._soup.select_one(
            f"div.shop-object div.form-label:-soup-contains('{label}')"
        )
        if label_node:
            return label_node.next_sibling
        return None

    def get_shop_field_value(self, label: str) -> str:
        node = self.get_shop_field_element(label)
        if node:
            return node.text
        return None

    def parse(self) -> Shop:
        name = self._soup.select_one("span.headline").text
        id = self.try_extract_number(self._soup.select_one("a.current").text)

        homepage = self.get_shop_field_value("Homepage")
        # TODO cloudflare email obfuscation
        contact = None  # self.get_shop_field_value("Contact")

        location_elem = self.get_shop_field_element("Location")
        location = self.try_get_text("span.shop-location", location_elem)
        shipping = self.try_get_text("small", location_elem)

        return Shop(
            id=id,
            name=name,
            homepage=homepage,
            contact=contact,
            location=location,
            shipping=shipping,
        )



import datetime
from tenji.model.user.profile import About, Profile
from tenji.parser.parser_base import ParserBase


class ProfileParser(ParserBase):
    def parse(self) -> Profile:
        username = self._soup.select_one("span.h1-headline-value > span.headline").text

        subtitle = self.try_get_text("span.h1-headline-value > span.subtitle")
        status = self.try_get_text("span.tbx-target-STATUS")

        # get background from style
        banner = self.try_get_style_background(
            self._soup.select_one("div.the-banner").get("style")
        )

        avatar = self._soup.select_one("img.the-avatar").get("src")
        stats_container = self._soup.select_one("div.object-stats")
        last_visit_node = stats_container.select_one("span:nth-child(1)")
        last_visit_relative = last_visit_node.text
        last_visit = self.try_parse_mfc_time(last_visit_node.get("title"), datetime.datetime.utcnow())
        joined_node = stats_container.select_one("span:nth-child(2)")
        joined_relative = joined_node.text
        joined = self.try_parse_mfc_time(joined_node.get("title"))

        eye_icon_node = stats_container.select_one("span.icon-eye")
        hits = self.try_extract_number(eye_icon_node.next_sibling)
        placement = self.try_extract_number(stats_container.select_one("small").text)

        about_container = self._soup.select_one("div.data_2")

        level = self.try_get_text(
            "div.form-label:-soup-contains('Level') + div > a", about_container
        )
        gender = self.try_get_text(
            "div.form-label:-soup-contains('Gender') + div > a", about_container
        )
        age = self.try_get_text(
            "div.form-label:-soup-contains('Age') + div", about_container
        )
        location = self.try_get_text(
            "div.form-label:-soup-contains('Location') + div", about_container
        )
        occupation = self.try_get_text(
            "div.form-label:-soup-contains('Occupation') + div", about_container
        )
        homepage = self.try_get_value(
            "div.form-label:-soup-contains('Homepage') + div > a", "href", about_container
        )
        shows = self.try_get_text(
            "div.form-label:-soup-contains('Shows(s)') + div > a", about_container
        )
        books = self.try_get_text(
            "div.form-label:-soup-contains('Book(s)') + div", about_container
        )
        games = self.try_get_text(
            "div.form-label:-soup-contains('Game(s)') + div", about_container
        )
        games = self.try_get_text(
            "div.form-label:-soup-contains('Game(s)') + div", about_container
        )
        moe_points = self.try_get_text(
            "div.form-label:-soup-contains('MOE Point(s)') + div", about_container
        )

        about = About(
            level=level,
            gender=gender,
            age=age,
            location=location,
            occupation=occupation,
            homepage=homepage,
            shows=shows,
            books=books,
            games=games,
            moe_points=moe_points,
        )

        return Profile(
            username=username,
            subtitle=subtitle,
            status=status,
            banner=banner,
            avatar=avatar,
            last_visit=last_visit,
            joined=joined,
            hits=hits,
            placement=placement,
            about=about,
        )



from tenji.model.user.collectionstatus import CollectionStatus
from tenji.request.request_base import RequestBase


class CollectionRequest(RequestBase):
    def __init__(
        self,
        username: str,
        status: CollectionStatus,
        page: int = 1,
    ) -> None:
        self.username = username
        self.status = status
        self.page = page

    def getPath(self) -> str:
        return f"{self.BASE_URL}users.v4.php?mode=view&username={self.username}&tab=collection&page={self.page}&status={self.status}&output=0"

    def getMethod(self):
        return "GET"



from typing import Optional
from pydantic import BaseModel
from datetime import datetime


class About(BaseModel):
    level: int
    gender: Optional[str] = None
    age: Optional[str] = None
    location: Optional[str] = None
    occupation: Optional[str] = None
    homepage: Optional[str] = None
    shows: Optional[str] = None
    books: Optional[str] = None
    games: Optional[str] = None
    music: Optional[str] = None
    moe_points: Optional[str] = None
    computer: Optional[str] = None
    camera: Optional[str] = None


class Profile(BaseModel):
    username: str

    subtitle: Optional[str] = None
    status: Optional[str] = None
    banner: Optional[str] = None
    avatar: Optional[str] = None
    rank: Optional[str] = None
    last_visit: datetime = None
    joined: datetime = None
    hits: Optional[int] = None
    placement: Optional[int] = None

    about: About



from tenji.model.category import get_item_category_from_str
from tenji.model.item.item import Item
from tenji.parser.parser_base import ParserBase


class ItemParser(ParserBase):
    def parse(self) -> Item:
        id = self.try_extract_number(self._soup.select_one("a.current").text)
        name = self._soup.select_one("span.h1-headline-value > span").text
        thumbnail = self._soup.select_one("span.h1-headline-icon > img").get("src")
        category = get_item_category_from_str(
            self.get_next_sibling_of("span.icon-tag").text
        )

        return Item(id=id, name=name, thumbnail=thumbnail, category=category)



from tenji.model.shop.shop_category import ShopCategory
from tenji.model.shop.shop_list_item import ShopListItem
from tenji.parser.parser_base import ParserBase


class ShopsParser(ParserBase):
    def parse(self) -> list[ShopListItem]:
        shops = []
        results = self._soup.select("#wide > div > section > div.results > div.result")

        for result in results:
            id = self.try_extract_number(result.select_one("a").attrs["href"])
            name = result.select_one("div.list-anchor a").text
            icon = result.select_one("img.list-icon").get("src")

            actions = result.select_one("div.list-actions")
            category = ShopCategory(
                self.try_extract_number(actions.select_one("a").attrs["href"][-3:])
            )
            location = self.try_get_text("a:nth-child(2)", actions)

            shops.append(
                ShopListItem(
                    id=id,
                    name=name,
                    icon=icon,
                    category=category,
                    location=location,
                )
            )

        return shops



from tenji.model.shop.shop_category import ShopCategory
from tenji.request.request_base import RequestBase


class ShopsRequest(RequestBase):
    def __init__(
        self,
        keywords: str = None,
        location: str = None,
        average_score: int = None,
        category: ShopCategory = None,
        page: int = 1,
    ) -> None:
        self.keywords = keywords
        self.location = location
        self.average_score = average_score
        self.category = category
        self.page = page

    def getPath(self) -> str:
        path = f"{self.BASE_URL}shops.v4.php"

        params = {
            "keywords": self.keywords,
            "location": self.location,
            "averageScore": self.average_score,
            "categoryId": self.category,
            "page": self.page,
        }

        return f"{path}?{self.build_params_url(params)}"

    def getMethod(self):
        return "GET"



from pydantic import BaseModel

from tenji.model.item.partner_status import PartnerListingStatus



class PartnerListing(BaseModel):
    jan: int
    shop_name: str
    shop_icon: str
    status: PartnerListingStatus
    url: str


from tenji.request.request_base import RequestBase


class ParserException(Exception):
    def __init__(self, message):
        super().__init__(message)

    @staticmethod
    def from_request(request: RequestBase, previous: Exception):
        msg = ParserException(f"Failed to parse {request.getPath()}")
        return ParserException(msg).with_traceback(previous.__traceback__)



from enum import IntEnum


class ItemCategory(IntEnum):
    Prepainted = 1
    ActionDolls = 2
    Trading = 3
    GarageKits = 4
    Plushes = 5
    Accessories = 6
    ModelKits = 10
    Linens = 13
    Dishes = 14
    HangedUp = 15
    Apparel = 17
    OnWalls = 18
    Stationeries = 20
    Misc = 16
    Books = 21
    Music = 22
    Video = 26
    Games = 28
    Software = 29


def get_item_category_from_str(category: str) -> ItemCategory:
    if category == "Prepainted":
        return ItemCategory.Prepainted
    elif category == "Action/Dolls":
        return ItemCategory.ActionDolls
    elif category == "Trading":
        return ItemCategory.Trading
    elif category == "Garage Kits":
        return ItemCategory.GarageKits
    elif category == "Plushes":
        return ItemCategory.Plushes
    elif category == "Accessories":
        return ItemCategory.Accessories
    elif category == "Linens":
        return ItemCategory.Linens
    elif category == "Dishes":
        return ItemCategory.Dishes
    elif category == "Hanged up":
        return ItemCategory.HangedUp
    elif category == "Apparel":
        return ItemCategory.Apparel
    elif category == "On Walls":
        return ItemCategory.OnWalls
    elif category == "Stationeries":
        return ItemCategory.Stationeries
    elif category == "Misc":
        return ItemCategory.Misc
    elif category == "Books":
        return ItemCategory.Books
    elif category == "Music":
        return ItemCategory.Music
    elif category == "Video":
        return ItemCategory.Video
    elif category == "Games":
        return ItemCategory.Games
    elif category == "Software":
        return ItemCategory.Software



import datetime
from typing import Optional
from pydantic import BaseModel

from tenji.model.item.item import Item
from tenji.model.paginated import Pagination


class UserList(BaseModel):
    id: int
    name: str
    owner: str
    icon: str
    created: datetime.datetime
    description: Optional[str] = None
    items: list[Item] = []
    tags: Optional[list[str]] = []
    pagination: Optional[Pagination] = None



import urllib.parse


class RequestBase:
    BASE_URL = "https://myfigurecollection.net/"

    def getPath(self):
        raise NotImplementedError()

    def getMethod(self):
        raise NotImplementedError()

    def build_params_url(self, params={}) -> str:
        params = {k: v for k, v in params.items() if v is not None}
        return urllib.parse.urlencode(params)
    
    def build_params(self, params={}):
        params = {k: v for k, v in params.items() if v is not None}
        return params



from http.cookies import SimpleCookie
from tenji.exceptions import *
from tenji.mfc_response import MFCResponse
from tenji.model import *
from tenji.parser import *
from tenji.request import *
import aiohttp
import logging

from tenji.request.request_base import RequestBase

class MFCException(Exception):
    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        return self.message


class MfcClient:
    USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/111.0.0.0 Safari/537.36"

    def __init__(self, session_id: str = None) -> None:
        cookies = {}
        if session_id:
            cookies["PHPSESSID"] = session_id

        self._session = aiohttp.ClientSession(
            headers={"User-Agent": self.USER_AGENT}, cookies=cookies
        )
        self.logger = logging.getLogger(__name__)

    def __exit__(self, exc_type, exc_value, traceback):
        self._session.close()

    def set_session(self, session: aiohttp.ClientSession):
        self._session = session

    async def is_logged_in(self) -> tuple[bool, str]:
        """Returns a tuple of whether the client is logged in and the username"""
        req = HomeRequest()
        res = await self.__perform_modeled_request(req)
        parser = HomeParser(res)
        meta = parser.parse()
        return (meta.is_guest, meta.username)

    async def login(self, username: str, password: str) -> tuple[bool, SimpleCookie]:
        """Logs in to MFC using the given username and password"""

        req = LoginRequest(username, password)
        response = await self.__perform_modeled_request(req)
        success = "Sorry, check your username and password" not in response.soup.text
        cookies = response.response.cookies
        if cookies:
            self._session.cookie_jar.update_cookies(cookies)
        filtered = self._session.cookie_jar.filter_cookies(
            "https://myfigurecollection.net"
        )
        return (success, filtered)

    async def logout(self) -> bool:
        url = "https://myfigurecollection.net/session/signout/"

        async with self._session.get(url) as response:
            if response.status != 200:
                raise Exception("Failed to perform request")
            html = await response.text()

        # TODO check if logout was successful

        return True

    async def get_profile(self, username: str) -> Profile:
        """Returns a Profile object for the given username"""
        req = ProfileRequest(username)
        res = await self.__perform_modeled_request(req)
        try:
            parser = ProfileParser(res)
            profile = parser.parse()
        except Exception as e:
            raise ParserException.from_request(req, e)
        return profile

    async def get_collection(
        self, username: str, status: Collection, page: int = 1
    ) -> Collection:
        """Returns a Collection object for the given username and status"""
        req = CollectionRequest(username, status, page)
        res = await self.__perform_modeled_request(req)
        try:
            parser = CollectionParser(res)
            collection = parser.parse()
        except Exception as e:
            raise ParserException.from_request(req, e)
        return collection

    async def get_lists(self, username: str) -> UserLists:
        """Gets public lists for a given user"""
        req = UserListsRequest(username)
        res = await self.__perform_modeled_request(req)
        try:
            parser = UserListsParser(res)
            lists = parser.parse()
        except Exception as e:
            raise ParserException.from_request(req, e)
        return lists

    async def get_item(self, id: int) -> Item:
        """Returns an Item object for the given id"""
        req = ItemRequest(id)
        res = await self.__perform_modeled_request(req)
        try:
            parser = ItemParser(res)
            item = parser.parse()
        except Exception as e:
            raise ParserException.from_request(req, e)

        return item

    async def get_list(self, id: int, page: int = 1) -> UserList:
        """Returns a List object for the given id"""
        req = UserListRequest(id, page)
        res = await self.__perform_modeled_request(req)
        try:
            parser = UserListParser(res)
            list = parser.parse()
        except Exception as e:
            raise ParserException.from_request(req, e)

        return list

    async def get_partner_listings(self, item_id: int) -> list[PartnerListing]:
        """Returns item availabilities from partners"""
        req = BuyItemRequest(item_id)
        res = await self.__perform_modeled_request(req)
        
        try:
            parser = PartnerItemListingParser(res)
            listings = parser.parse()
        except Exception as e:
            raise ParserException.from_request(req, e)
        return listings

    async def get_shop(self, id: int) -> Shop:
        """Returns a Shop object for the given id"""
        req = ShopRequest(id)
        res = await self.__perform_modeled_request(req)
        try:
            parser = ShopParser(res)
            shop = parser.parse()
        except Exception as e:
            raise ParserException.from_request(req, e)
        return shop

    async def get_shops(
        self,
        keywords: str = None,
        location: str = None,
        average_score: int = None,
        category: str = None,
        page: int = 1,
    ) -> list[ShopListItem]:
        """Returns a list of Shops"""
        req = ShopsRequest(keywords, location, average_score, category, page)
        res = await self.__perform_modeled_request(req)
        try:
            parser = ShopsParser(res)
            shop = parser.parse()
        except Exception as e:
            raise ParserException.from_request(req, e)
        return shop

    async def __perform_modeled_request(self, req: RequestBase) -> MFCResponse:
        """Performs a request and returns the response"""

        method = req.getMethod()
        if method == "GET":
            async with self._session.get(req.getPath()) as response:
                if response.status != 200:
                    raise Exception(f"Failed to perform request: {req.getPath()}")
                response_body = await response.text()
        elif method == "POST":
            async with self._session.post(
                req.getPath(), data=req.getParams()
            ) as response:
                if response.status != 200:
                    raise Exception(f"Failed to perform request {req.getPath()}")
                response_body = await response.text()
  
        res = MFCResponse(response_body)
        return res



from enum import Enum

class PartnerListingStatus(Enum):
    MAYBE_AVAILABLE = "Maybe available"
    AVAILABLE = "Available"
    NOT_AVAILABLE = "Not available"
    
    def __str__(self):
        return self.value


from tenji.request.request_base import RequestBase


class ItemRequest(RequestBase):
    def __init__(self, id: int) -> None:
        self.id = id

    def getPath(self) -> str:
        return f"{self.BASE_URL}item/{self.id}"

    def getMethod(self):
        return "GET"



from tenji.request.request_base import RequestBase


class UserListsRequest(RequestBase):
    def __init__(self, username: str, page: int = 1) -> None:
        self.username = username
        self.page = page

    def getPath(self) -> str:
        return f"{self.BASE_URL}users.v4.php?mode=view&username={self.username}&tab=lists&page={self.page}"

    def getMethod(self):
        return "GET"



from tenji.model.category import get_item_category_from_str
from tenji.model.item.item import Item

from tenji.model.user.user_list import UserList
from tenji.parser.parser_base import ParserBase


class UserListParser(ParserBase):
    def parse(self) -> UserList:
        items = []

        icon = self._soup.select_one("span.h1-headline-icon > img").get("src")
        name = self._soup.select_one("span.h1-headline-value > span").text

        list_id = self.try_extract_number(self._soup.select_one("a.current").text)

        list_info_container = self._soup.select_one("div.itemlist-object")

        owner = list_info_container.select_one("a.user-anchor").text

        description = list_info_container.select_one("div.bbcode").decode_contents()
        created = self.try_parse_mfc_time(
            list_info_container.select_one("span.time > span:nth-child(2)").get("title")
        )

        item_stamps = self._soup.select("div.item-stamp")
        for item_stamp in item_stamps:
            link = item_stamp.select_one("a")
            item_id = self.try_extract_number(link.get("href"))
            item_thumbnail = link.select_one("img").get("src")
            item_name = link.select_one("img").get("alt")
            item_category = get_item_category_from_str(
                item_stamp.select_one("div.stamp-category").text
            )

            item = Item(
                id=item_id,
                name=item_name,
                thumbnail=item_thumbnail,
                category=item_category,
            )

            items.append(item)

        tags = []
        tag_nodes = self._soup.select("div.object-tag")
        for tag_node in tag_nodes:
            tag = tag_node.select_one("a").text
            tags.append(tag)

        pagination = self.try_parse_pagination()

        return UserList(
            id=list_id,
            name=name,
            owner=owner,
            icon=icon,
            created=created,
            description=description,
            items=items,
            tags=tags,
            pagination=pagination,
        )



from tenji.model.meta import Meta
from tenji.parser.parser_base import ParserBase


class HomeParser(ParserBase):
    def parse(self) -> Meta:
        user_menu = self._soup.select_one("div.user-menu")

        handle_link = user_menu.select_one("a.handle")

        is_guest = True

        notif_count = None  # anything greater than 9 will show as 9+
        username = None
        avatar = None

        if self._soup.select_one("span.icon-sliders") is None:
            is_guest = False

        if handle_link:
            notif_count = self.try_get_text("span.count", user_menu, None)
            username = self.try_get_text("span.username", handle_link)
            avatar = self.try_get_value("img", "src", handle_link)

        return Meta(is_guest=is_guest, username=username, avatar=avatar)



from typing import Optional
from pydantic import BaseModel

from tenji.model.shop.shop_category import ShopCategory


class ShopListItem(BaseModel):
    id: int
    name: str
    icon: str
    category: ShopCategory
    location: Optional[str]



from tenji.request import RequestBase


class HomeRequest(RequestBase):
    def __init__(self) -> None:
        pass

    def getPath(self) -> str:
        return self.BASE_URL

    def getMethod(self):
        return "GET"



from typing import Optional
from pydantic import BaseModel


class Shop(BaseModel):
    id: int
    name: str
    homepage: Optional[str]
    contact: Optional[str]
    location: Optional[str]
    shipping: Optional[str]



from tenji.model.item import PartnerListing
from tenji.parser.parser_base import ParserBase


class PartnerItemListingParser(ParserBase):

    def parse(self) -> list[PartnerListing]:
        self.parse_html_from_json(self.response.body)

        listings = []
        
        results = self._soup.select("div.result")
        for result in results:
            stamp = result.select_one("div.stamp")
            link = stamp.select_one("a")
            icon = link.select_one("img")["src"]
            url = link["href"]

            jan = self.try_get_url_query_value(url, "jan")

            stamp_data = stamp.select_one("div.stamp-data")
            name = stamp_data.select_one("div.stamp-anchor").text

            availability = stamp_data.select_one("div.item-availability").text

            listing = PartnerListing(
                shop_icon=icon,
                shop_name=name,
                url=url,
                availability=availability,
                jan=jan,
                status=availability
            )

            listings.append(listing)

        return listings




        


from tenji.request import RequestBase


class LoginRequest(RequestBase):
    def __init__(self, username: str, password: str) -> None:
        self.username = username
        self.password = password

    def getPath(self) -> str:
        return f"{self.BASE_URL}sessions.v4.php"

    def getMethod(self):
        return "POST"

    def getParams(self):
        return {
            "username": self.username,
            "password": self.password,
            "commit": "signIn",
            "from": f"{self.BASE_URL}session/signin/",
        }



from tenji.request.request_base import RequestBase


class UserListRequest(RequestBase):
    def __init__(self, id: int, page: int = 1) -> None:
        self.id = id
        self.page = page

    def getPath(self) -> str:
        return f"{self.BASE_URL}itemlists.v4.php?mode=view&itemListId={self.id}&tab=view&output=0&page={self.page}"

    def getMethod(self):
        return "GET"



from tenji.model.user.user_list_item import UserListItem
from tenji.model.user.users_lists import UserLists


from tenji.parser.parser_base import ParserBase


class UserListsParser(ParserBase):
    def parse(self) -> UserListItem:
        items = []

        results = self._soup.select("div.list-data")
        for result in results:
            link = result.select_one("div.list-anchor > a")
            name = link.text
            list_id = self.try_extract_number(link["href"])
            icon = result.select_one("img.list-icon")["src"]
            owner = result.select_one("div.list-meta > a").text

            creation_node = result.select_one("div.list-stats > span")
            created = self.try_parse_mfc_time(creation_node.get("title"))
            count = self.try_extract_number(creation_node.next_sibling.text)

            list = UserListItem(
                id=list_id,
                icon=icon,
                name=name,
                owner=owner,
                created=created,
                count=count,
            )

            items.append(list)

        pagination = self.try_parse_pagination()
        return UserLists(items=items, pagination=pagination)



from pydantic import BaseModel
from tenji.model.paginated import Pagination

from tenji.model.user.user_list_item import UserListItem


class UserLists(BaseModel):
    items: list[UserListItem] = []
    pagination: Pagination = None



from tenji.model.category import ItemCategory, get_item_category_from_str
from tenji.model.user.collection import Collection
from tenji.model.item.item import Item
from tenji.parser.parser_base import ParserBase


class CollectionParser(ParserBase):
    def parse(self) -> Collection:
        items = []

        results = self._soup.select("div.results > div.result")

        # TODO check if collection is empty
        for result in results:
            stamp = result.select_one("div.stamp")
            link = stamp.select_one("a")
            id = self.try_extract_number(link.get("href"))
            thumbnail = link.select_one("img").get("src")
            name = link.select_one("img").get("alt")

            cat_node = result.select_one("div.stamp-category")
            cat_id = self.try_extract_number(cat_node.get("class")[-1])
            category = ItemCategory(cat_id)

            item = Item(
                id=id,
                name=name,
                thumbnail=thumbnail,
                category=category,
            )

            items.append(item)

        pagination = self.try_parse_pagination()
        return Collection(items=items, pagination=pagination)



import datetime
from bs4 import BeautifulSoup, Tag
from urllib.parse import urlparse
from urllib.parse import parse_qs

import json
import re
from tenji.mfc_response import MFCResponse

from tenji.model.paginated import Pagination


class ParserBase:

    def __init__(self, response: MFCResponse) -> None:
        self.response = response
        self.__soup = None

    def parse_html(self, html: str):
        self.__soup = BeautifulSoup(html, "html.parser")
    
    def parse_json(self, json_str: str) -> dict:
        return json.loads(json_str)

    def parse_html_from_json(self, json_str: str) -> str:
        json_dict = self.parse_json(json_str)
        html_values = json_dict["htmlValues"]
        if html_values is None:
            return None
        
        html = html_values["WINDOW"]
        if html is None:
            return None
        
        self.parse_html(html)

    @property
    def _soup(self) -> BeautifulSoup:
        if self.__soup is None:
            self.parse_html(self.response.body)
        return self.__soup

    def try_get_tag(
        self,
        selector: str,
        parent: Tag = None,
    ):
        p = parent if parent else self._soup
        node = p.select_one(selector) if selector else p
        return node

    def try_get_text(self, selector: str, parent: Tag = None, default_value=None):
        node = self.try_get_tag(selector, parent)
        if node:
            return node.text
        return default_value

    def try_get_value(
        self, selector: str, attr: str, parent: Tag = None, default_value=None
    ):
        node = self.try_get_tag(selector, parent)
        if node and attr in node.attrs:
            return node.get(attr)
        return default_value

    def try_get_list(self, selector: str, parent: Tag = None, default_value=None):
        text = self.try_get_text(selector, parent, default_value)
        if text:
            return [x.strip() for x in text.split(",")]
        return default_value

    def try_extract_number(self, text: str, default_value=None):
        text = text.replace(",", "")
        match = re.search(r"\d+", text)
        if match:
            return int(match.group(0))
        return default_value

    def try_get_style_background(self, style: str):
        match = re.search(r"url\((.+)\)", style)
        if match:
            return match.group(1)
        return None

    def try_get_text(self, selector: str, parent: Tag = None, default_value=None):
        node = self.try_get_tag(selector, parent)
        if node:
            return node.text
        return default_value

    def get_next_sibling_of(self, selector: str, parent: Tag = None):
        node = self.try_get_tag(selector, parent)
        if node and node.next_sibling:
            return node.next_sibling
        return None

    def try_parse_mfc_time(self, date: str, default: datetime.datetime = None) -> datetime.datetime:
        try:
            # ex 12/21/2017, 13:01:50
            format = "%m/%d/%Y, %H:%M:%S"
            d = datetime.datetime.strptime(date, format)
            return d.replace(tzinfo=datetime.timezone.utc)
        except:
            return default

    def get_trailing_number(self, text: str) -> int:
        match = re.search(r"\d+$", text)
        if match:
            return int(match.group(0))
        return None
    
    def get_item_id_from_thumbnail(self, thumbnail: str) -> int:
        # https://static.myfigurecollection.net/upload/items/0/198579-4200e.jpg
        filename =  thumbnail.split("/")[-1]
        id = filename.split("-")[0]
        return int(id)
    
    def try_get_url_query_value(self, url: str, key: str, default_value=None):
        query = urlparse(url).query
        if query:
            params = parse_qs(query)
            if key in params:
                return params[key][0]
        return default_value

    def try_parse_pagination(self, parent: Tag = None) -> Pagination:
        p = parent if parent else self._soup
        pagination_controls = p.select_one("div.results-count")
        if pagination_controls is None:
            return None

        total_items = self.try_extract_number(
            pagination_controls.select_one("div.results-count-value").text
        )

        nav_controls = pagination_controls.select_one("div.results-count-pages")

        if nav_controls is None:
            return Pagination(
                current_page=1,
                has_next_page=False,
                total_pages=1,
                total_items=total_items,
            )

        current_link = nav_controls.select_one("a.nav-current")
        next_link = nav_controls.select_one("a.nav-next")
        last_link = nav_controls.select_one("a.nav-last.nav-end")

        current_page = self.try_extract_number(current_link.text)

        if last_link:
            total_pages = parse_qs(urlparse(last_link.get("href")).query)["page"][0]
        else:
            total_pages = current_page

        pagination = Pagination(
            current_page=current_page,
            has_next_page=next_link is not None,
            total_pages=total_pages,
            total_items=total_items,
        )

        return pagination



from typing import Optional
from pydantic import BaseModel

from tenji.model.item.item import Item
from tenji.model.paginated import Pagination


class Collection(BaseModel):
    items: list[Item] = []
    pagination: Optional[Pagination] = None
